structure SymbolTable (* :> SYMBOL_TABLE *) =
  struct
    exception Subscript

    type Index = string
    type Dimensions = Index list
    (* type Traversal *)
    type 'a T = 'a TernaryTree.T
    type 'a Region = {base : 'a T, dimensions : Dimensions option}
    
    fun symbol name =
          name

    fun name theName =
          theName
    
    (* The rest implements symbol tables as
       ternary string trees *)
    fun empty () =
          TernaryTree.empty ()

    fun insert aTable name binding =
          TernaryTree.insert aTable name binding

    fun delete aTable name =
          TernaryTree.delete aTable name

    fun lookup aTable name =
          TernaryTree.lookup aTable name

    fun tabulate (indices, tabF) =
          foldl (fn (index, acc) => insert acc index (tabF index))
                (empty ()) indices

    fun create (indices, value) =
          tabulate (indices, fn _ => value)
    
    fun sub (aTable, index) =
          case lookup aTable index of
            NONE =>
              raise Subscript
          | SOME v =>
              v

    fun dimensions aTable =
          TernaryTree.dimensions aTable
          
    exception InternalError

    fun mapiRegionF f base (index, acc) =
          case lookup base index of
            NONE =>
              raise InternalError
          | SOME v =>
              insert acc index (f (index, v))

    fun mapiRegion f {base, dimensions} =
          foldl (mapiRegionF f base) (empty ()) dimensions

    fun mapi f {base, dimensions = NONE} =
          mapiRegion f {base = base, dimensions = dimensions base}
      | mapi f {base, dimensions = SOME dims} =
          mapiRegion f {base = base, dimensions = dims}

    fun map f aTable =
          mapi (fn (_, v) => f v)
               {base = aTable, dimensions = NONE}

    fun appendF fromTable (addIndex, accTable) =
          case lookup fromTable addIndex of
            NONE =>
              raise InternalError
          | SOME v =>
              insert accTable addIndex v

    fun append (aTable, addTable) =
          let
            val addIndices = dimensions addTable
          in
            foldl (appendF addTable) aTable addIndices
          end
  end

(*
functor FTestSymbolTable(s : SYMBOL_TABLE) =
  struct
    fun testIt () =
          let
            val a = s.symbol "test"
            val b = s.symbol "b"
            val t = s.empty
            val t1 = s.insert t a "this is the value of symbol a"
            val t2 = s.insert t1 b "this is the binding of symbol b"
          in
            (s.lookup t a) = NONE andalso
            (s.lookup t1 a) = SOME("this is the value of symbol a") andalso
            (s.lookup t1 b) = NONE andalso
            (s.lookup t2 b) = SOME("this is the binding of symbol b") andalso
            (s.lookup t2 a) = SOME("this is the value of symbol a")
          end
  end

structure TestSymbolTable = FTestSymbolTable(SymbolTable) *)

(* val _ = if (TestSymbolTable.testIt ()) then print "OK!" else (); *)