(* This file is part of ML Abstract UI - a user interface toolkit for SML'97.
   Copyright (C) 1999-2002 ANOQ of the Sun (alias Johnny Andersen)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   I reserve the right to alter the license in future versions of ML Abstract UI.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   You can get this software from http://www.HardcoreProcessing.com/
   E-mail: anoq@HardcoreProcessing.com
*)

structure TestUtil (* : TEST_UTIL *) =
  struct
    val output = print

    (* This module has some ugly imperative stuff because we want
       to keep the _interface_ to this module very simple
       - to encourage as much testing as possible :) *)
    val verbose = ref(true)
    val indent = ref nil : string list ref
  
    fun setVerbose value =
          verbose := value

    fun indentedOutput msg =
          if !verbose then
            (List.app output (!indent);
             output msg; output "\n")
          else
            ()

    fun outputBegin msg =
          (indentedOutput msg;
           indent := ("  " :: (!indent)))

    fun outputEnd msg =
          (indent :=
             (case !indent of
                nil =>
                  (output "Error in FTestUtil.outputEnd - ";
                   output "probably unbalanced calls to ";
                   output "outputBegin and outputEnd";
                   nil)
              | (_::rest) =>
                  rest);
           indentedOutput msg)
    
    fun doTest nil =
          true
      | doTest ((name, f)::fList) =
          let
            val _ = outputBegin ("Testing " ^ name ^ ":")

            val error =
                  if f () then
                    NONE
                  else
                    SOME ("Found error in: " ^ name)
                      handle _ =>
                        (SOME ("Found error (exception raised) in: " ^ name))
          in
            case error of
              NONE =>
                (outputEnd (name ^ " OK!");
                 doTest fList)
            | SOME msg =>
                (outputEnd msg;
                 false)
          end
  end
